# encoding: utf-8
"""
@author: BrikerMan
@contact: eliyar917@gmail.com
@blog: https://eliyar.biz

@version: 1.0
@license: Apache Licence
@file: __init__.py.py
@time: 2019-01-19 11:49

"""
from .base_model import ClassificationModel
from .blstm_model import BLSTMModel
from .cnn_lstm_model import CNNLSTMModel
from .cnn_model import CNNModel

if __name__ == "__main__":
    print("Hello world")
